/*
 * jpgQ - Qurey about jpeg images
 *   by uratan! 2018.6.19
 */
	/*
	 *  about JPEG
	 * https://www.setsuki.com/hsp/ext/jpg.htm
	 *  about sub-sampling
	 * https://nixeneko.hatenablog.com/entry/2017/09/23/000000
	 * http://animencode.hatenablog.com/entry/2014/09/20/014902
	 */
#include <stdio.h>
#include <stdlib.h>

//#define STDERR stdout		// for test
#define STDERR stderr

/*
 *
 */
int main(int argc, char *argv[]);
int decode(char *argv);
int get_value(int bytes);

/*
 *
 */
char *fn;		/* current filename */
FILE *fp;		/* fp for above file */
int opt_no_fn;	/* -x: omit filename display */
				/*   (force $1 as X-size whether including SPC or not) */

/*
 *
 */
int main(int argc, char *argv[])
{
	char *prog;

#ifdef __WIN32__
	extern void xargs(int *argc, char ***argv);	/* in xargs2.c */
	xargs(&argc, &argv);
#endif /* __WIN32__ */

	prog = argv[0];

	opt_no_fn = 0;
	if(argc >= 2) {
		if((argv[1][0] == '-') && (argv[1][1] == 'x')) {
			opt_no_fn = 1;
			--argc;
			++argv;
		}
	}

	if(argc < 2) {
		fprintf(STDERR, "usage: %s [-x] files ... \n", prog);
		exit(1);
	}

	while(--argc) {
		decode(*++argv);
	}

	exit(0);
}

/*
 *
 */
int decode(char *argv)
{
	int val, i;
	int gotten;
	int marker, seg_len;
	int width, height, Q, frames, progressive, subsampling;

	fn = argv;
	fp = fopen(fn, "rb");
	if(fp == NULL) {
		fprintf(STDERR, "can't open file: %s\n", fn);
		return(1);
	}

	/* check SOI (0xFFD8) */
	val = get_value(2);
	if(val != 0xffd8) {
		fprintf(STDERR, "non JPEG file: %s\n", fn);
		goto close_end;
	}

#define GotNONE   (0)
#define GotDQT    (1<<0)
#define GotSOFx   (1<<1)
#define GotALL   ((1<<2)-1)
	gotten = GotNONE;

	while(gotten != GotALL) {
		/*
		 * get segment marker
		 */
		val = get_value(2);
		if((val & 0xff00) != 0xff00) {
			fprintf(STDERR, "invalid segment found: %s\n", fn);
			goto close_end;
		}
		marker = val;

		/*
		 * get segment length
		 */
		val = get_value(2);
		seg_len = val - 2;		/* decrease myself */

		switch(marker) {
		case 0xffdb:				/* DQT */
				val = get_value(1);
				if(val != 0x00) {
					seg_len -= 1;
					break;
				}

#define SUMof_Qorg  3688			/* see bottom of this file */

				val = 0;
				for(i=0; i<64; i++) {
					val += get_value(1);
				}

#define PREC 100					/* fixed point, precision: n.00 */
				if(val > SUMof_Qorg) {
					Q = (50*PREC * SUMof_Qorg) / val;
				} else {
					Q = 100*PREC - (50*PREC * val) / SUMof_Qorg;
				}
				Q = (Q + 88) / PREC;			/* 88 is cut & try */
				gotten |= GotDQT;

				seg_len -= 1 + 64;
				break;

		case 0xffc0:					/* SOF0 */
		case 0xffc1:					/* SOF1 */
		case 0xffc2:					/* SOF2 */
		case 0xffc3:					/* SOF3 */
				progressive = (marker == 0xffc2) ? 1 : 0;
				val = get_value(1);		/* P - Sample Precision : Tv̐x*/
				val = get_value(2);		/* Y - Number of Lines : 摜̍ */
				height = val;
				val = get_value(2);		/* X - Samples Per Line : 摜̉ */
				width = val;
				val = get_value(1);		/* Nf - Number of Image Componets in Frame */
				frames = val;
				gotten |= GotSOFx;
				seg_len -= 6;

				for(i=0; i<1 /*frames*/; i++) {
					val = get_value(1);		/* Ci - \vfʎq */
					val = get_value(1);		/* Hi, Vi - Tvv */
					subsampling = val;
					val = get_value(1);		/* Tqi - ʎqe[u̎ʎq */
					seg_len -= 3;
				}
				break;

		default:
				break;
		}

		/* skip this segment or rest of the segment */
		fseek(fp, seg_len, SEEK_CUR);
	}

	if(gotten == GotALL) {
		long len;
		double kb_Mpx;		/* kByes / Mega-pixel */

		fseek(fp, 0, SEEK_END);
		len = ftell(fp);
		if((width != 0) && (height != 0)) {
			kb_Mpx = len / 1024.0 / (width * height / 1000.0 / 1000.0);
		} else {
			kb_Mpx = 0.0;
		}

		if(opt_no_fn) {
			printf(   "%d %d -q %d",     width, height, Q);
		} else {
			printf("%s %d %d -q %d", fn, width, height, Q);
		}

		printf("  %7.2f", kb_Mpx);

		if(frames == 1) {
			printf(" grayscale");
		}
		else {
			switch(subsampling) {
			case 0x11: printf(" ss444"); break;
			case 0x12: printf(" ss440"); break;
			case 0x21: printf(" ss422"); break;
			case 0x22: printf(" ss420"); break;
			case 0x41: printf(" ss411"); break;
			case 0x42: printf(" ss410"); break;
			default:   printf(" ssX%02x", subsampling); break;
			}
					/*
					 *  Y   Cb  Cr | ݒ
					 * ------------+-------
					 *  11  11  11 | 4:4:4
					 *  12  11  11 | 4:4:0
					 *  21  11  11 | 4:2:2
					 *  22  11  11 | 4:2:0
					 *  41  11  11 | 4:1:1
					 *  42  11  11 | 4:1:0
					 */
		}

		if(progressive == 1) {
			printf(" progressive\n");
		} else {
			printf("\n");
		}
	}

close_end:
	if(fp != NULL) {
		fclose(fp);
	}

	return(0);
}

/*
 *
 */
int get_value(int bytes)
{
	int val, i, c;

	val = 0;
	for(i=0; i<bytes; i++) {
		c = fgetc(fp);
		if(c == EOF) {
			fprintf(STDERR, "unexpected EOF: %s\n", fn);
			return 0x88888888;
		}

		val = (val << 8) + c;
	}

	return val;
}

#if 0000
/*
 * http://d.hatena.ne.jp/Blue-ringed-Octopus/20100629
 *    
 * http://web.archive.org/web/20060716124317/www02.so-net.ne.jp/~koujin/jpeg/DCT2DQ.html
 *
 * ̃AvbgŌɂĂʎqe[u ISO/IEC 10918-1, ITU-T T.81 
 * Pxp̃e[uƂė Ă̂łB
 * (Qorg)ij =
 *   16  11  10  16  24  40  51  61
 *   12  12  14  19  26  58  60  55
 *   14  13  16  24  40  57  69  56
 *   14  17  22  29  51  87  80  62
 *   18  22  37  56  68 109 103  77
 *   24  35  55  64  81 104 113  92
 *   49  64  78  87 103 121 120 101
 *   72  92  95  98 112 100 103  99   ... all sum is: 3688
 */
#endif /* 0000 */

#ifdef __WIN32__
/*
 * xargs2.c - porting xargs1.c for Borland Free C compiler version 5.5.1
 *   by uratan! 2003.12.20
 */
/*
 * additional codes are ifdef'ed by '__WIN32__'
 */

#ifdef __WIN32__ /* =================================== */
/*
 * at this port, directory and hidden file are matched by default
 * because of specification of _findfirst(), _findnext()
 *
 * also these compile options are changed from them of my MS-C V6 era.
 */

#define USE_CFG 0		/* may be needless */
#define USE_CCL 1		/* must be 1 */
#define DOT_DLM 0		/* for matching 'xxx.c.bak' */

#define CVLOWER	0	/* if 1, wildargs to lower */

#endif /* __WIN32__ =================================== */

/* wild card expand routine
 *  for Turbo-C, MS-C, LSI-C..
 *
 *  This code is in the public domain.
 *
 *  Bug for use of wild card in directory part was fixed.
 *
 * $Id: xargs1.c 1.9a 92/05/03 06:00:00 AssistantIO Exp $
 *
 *  Bug for DOT_DLM=0 was fixed.
 *
 * $Id: xargs1.c 1.9 92/03/16 16:30:57 AssistantIO Exp $
 *
 *  Buggy code was fixed, macro became externaly definable.
 *
 * $Id: xargs1.c 1.8 91/11/24 08:15:12 AssistantIO Exp $
 *
 *  Configuration from file was added.
 *
 * $Id: xargs1.c 1.7 91/10/27 17:15:12 AssistantIO Exp $
 *
 *  Definition of ustrcmp() was changed.
 *
 * $Id: xargs1.c 1.6 91/08/31 09:15:12 AssistantIO Exp $
 *
 *  Recursive globbing was added.
 *  Use of character class regexp was added.
 *
 * $Id: xargs1.c 1.5 91/04/17 08:51:50 AssistantIO Exp $
 */
#ifndef USE_CFG
#define USE_CFG 1		/* if 1, configuratin from file */
#endif
#ifndef USE_CCL
#define USE_CCL 1		/* if 0, to the original source */
#endif
#ifndef DOT_DLM
#define DOT_DLM 1		/* if 0, '.' is not delimiter */
#endif

/* wild card expand routine
 *  for Turbo-C, MS-C, LSI-C..
 *
 * $Id: xargs.c 1.2 90/12/28 21:51:50 serow Exp Locker: serow $
 */

#include <stdio.h>
#include <dos.h>
#include <string.h>
#include <io.h>
#include <stdlib.h>

#define PSP	_psp

#ifndef ArgsAttr
#define ArgsAttr	1	/* Default attribute when file search */
#endif
#ifndef DirAttr
#define DirAttr		16	/* Default attribute when dir search */
#endif
#ifndef DO_SORT
#define DO_SORT	1	/* NON 0:wild card args will be sorted by name */
#endif
#ifndef CVLOWER
#define CVLOWER	1	/* if 1, wildargs to lower */
#endif

#if USE_CCL
#define LBRACE	0x7f
#if DOT_DLM
#define is_delimiter(c)	((c) == '\0' || (c) == '.')
#else
#define is_delimiter(c)	((c) == '\0')
#endif

#define IGNORE_CASE	1

#define isupper(c)	((c) >= 'A' && (c) <= 'Z')
#define tolower(c)	((c) | 0x20)
#endif /* USE_CCL */

#define isspace(c)	((c) == ' ' || (c) == '\t' || (c) == '\n' )
#define UCH(c)  ((unsigned char)(c))
#define isk1(c) ((0x81<=UCH(c)&&UCH(c)<=0x9F)||(0xE0<=UCH(c)&&UCH(c)<=0xFC))
#define isk2(c) ((0x40<=UCH(c)&&UCH(c)<=0x7E)||(0x80<=UCH(c)&&UCH(c)<=0xFC))

#define REALLOC_UNIT	128

#ifdef LDATA
#undef LDATA
#endif

#ifdef __TURBOC__
#if (sizeof(char *) == 4)
#define LDATA	1
#endif
#else /* may be MS-C */
#if defined(M_I86LM) || defined(M_I86CM) || defined(M_I86HM)
#define LDATA	1
#endif
#endif

#if defined(LSI_C)
/* union REGS dose not have '.cflag' but '.flags' */
#else
#define HAVE_CFLAG	1
#endif

#ifndef __WIN32__ /* =================================== */

#ifdef FP_OFF
#undef FP_OFF
#endif
#ifdef FP_SEG
#undef FP_SEG
#endif
#ifdef MK_FP
#undef MK_FP
#endif
#define FP_OFF(fp) ((unsigned int)(fp))
#define FP_SEG(fp) ((unsigned int)((unsigned long)(fp) >> 16))
#define MK_FP(sg,of) ((void far *)(((unsigned long)(sg) << 16)|(unsigned)(of)))

typedef struct {
	char           d_buf[21];
	char           d_attribute;
	unsigned short d_time;
	unsigned short d_date;
	long           d_size;
	char           d_name[13];
} DTA_BUF;

#if !USE_CCL
static DTA_BUF    dtabuf;
#endif
static union REGS reg, nreg;

#ifdef LDATA
static struct SREGS sreg;
#endif

#endif /* __WIN32__ =================================== */

static char     ERNEM[] = "not enought mem\r\n";
/***************************************************************/

#if USE_CCL
static int regcmp(char *t, char *p)
{
	int r, include, exclude;
	unsigned int ks, ke, kp, kt;
	char *s, *u;

	while (*p) {
		switch (*p) {
		case LBRACE:
			if (is_delimiter(*t)) {
				return 0;
			}
			s = t;
			u = s;
			p++;
			do {
				t = s;
				r = 1;
				do {
					if (is_delimiter(*p)) {
						return -1;
					}
					kt = UCH(*t++);
					if (isk1(kt) && isk2(*t)) {
						kt = (kt << 8) + UCH(*t++);
					}
					kp = UCH(*p++);
					if (isk1(kp) && isk2(*p)) {
						kp = (kp << 8) + UCH(*p++);
					}
#if IGNORE_CASE
					if (isupper(kt)) {
						kt = tolower(kt);
					}
					if (isupper(kp)) {
						kp = tolower(kp);
					}
#endif
					if (kt != kp) {
						r = 0;
					}
				} while (*p != ',' && *p != '}');
				if (r == 1 && t - s > u - s) {
					u = t;
				}
				if (*p == ',') {
					p++;
				}
			} while (*p != '}');
			if (u == s) {
				return 0;
			}
			p++;
			t = u;
			break;
		case '?':
			if (is_delimiter(*t)) {
				return 0;
			}
			if (isk1(*t) && isk2(*(t + 1))) {
				t++;
			}
			t++;
			p++;
			break;
		case '*':
			while (!((r = regcmp(t, p + 1)) != 0
					 || is_delimiter(*t))) {
				if (isk1(*t) && isk2(*(t + 1))) {
					t++;
				}
				t++;
			}
			return r;
		case '[':
			if (is_delimiter(*t)) {
				return 0;
			}
			include = 0;
			exclude = 0;
			p++;
			if (*p == '^') {
				exclude = 1;
				p++;
			}
			kt = UCH(*t++);
			if (isk1(kt) && isk2(*t)) {
				kt = (kt << 8) + UCH(*t++);
			}
#if IGNORE_CASE
			if (isupper(kt)) {
				kt = tolower(kt);
			}
#endif
			do {
				if (is_delimiter(*p)) {
					return -1;
				}
				ks = UCH(*p++);
				if (isk1(ks) && isk2(*p)) {
					ks = (ks << 8) + UCH(*p++);
				}
				ke = ks;
				if (*p == '-' && *(p + 1) != ']') {
					p++;
					if (is_delimiter(*p)) {
						return -1;
					}
					ke = UCH(*p++);
					if (isk1(ke) && isk2(*p)) {
						ke = (ke << 8) + UCH(*p++);
					}
				}
#if IGNORE_CASE
				if (isupper(ks)) {
					ks = tolower(ks);
				}
				if (isupper(ke)) {
					ke = tolower(ke);
				}
#endif
				if (kt >= ks && kt <= ke) {
					include = 1;
				}
			} while (*p != ']');
			if (include == exclude) {
				return 0;
			}
			p++;
			break;
#if DOT_DLM
		case '.':
			if (!(is_delimiter(*t))) {
				return 0;
			}
			if (*t == '.') {
				t++;
			}
			p++;
			break;
#endif
		default:
			kt = UCH(*t++);
			if (isk1(kt) && isk2(*t)) {
				kt = (kt << 8) + UCH(*t++);
			}
			kp = UCH(*p++);
			if (isk1(kp) && isk2(*p)) {
				kp = (kp << 8) + UCH(*p++);
			}
#if IGNORE_CASE
			if (isupper(kt)) {
				kt = tolower(kt);
			}
			if (isupper(kp)) {
				kp = tolower(kp);
			}
#endif
			if (kt != kp) {
				return 0;
			}
		}
	}
	return *t == '\0' ? 1 : 0;
}
#endif /* USE_CCL */

static void error(char * s)
{
	write(2, "xargs: ", 7);
	write(2, s, strlen(s));
	exit(1);
}

#ifdef __WIN32__ /* =================================== */

static char *find_file(char *dir, int attr, struct _finddata_t *fdt)
{
	static long ff_handle[64];		/* nest depth */
	static int ff_index = -1;
	long ret;

	if (dir != (char *) 0) { /* get first entry */
		/*
		 * attr is ignored, cause almost all element will be match
		 */
		ff_handle[++ff_index] = _findfirst(dir, fdt);
		if(ff_handle[ff_index] != -1) {
			ret = 0;	/* success */
		} else {
			ff_index--;
			ret = 1;	/* fail */
		}
	} else {			/* get next entry */
		ret = _findnext(ff_handle[ff_index], fdt);
		if(ret)
			_findclose(ff_handle[ff_index--]);
	}

	if(ret)
		return (char *) 0;

	if (fdt->name[0] == 5) fdt->name[0] = 0xE5;
	return fdt->name;
}

#else /* __WIN32__ =================================== */

#if USE_CCL

static void set_dta(DTA_BUF *dtabuf)
{
	reg.h.ah = 0x1A;
#ifdef LDATA
	reg.x.dx = FP_OFF(dtabuf);
	sreg.ds = FP_SEG(dtabuf);
	intdosx(&reg, &nreg, &sreg);
#else
	reg.x.dx = (unsigned) dtabuf;
	intdos(&reg, &nreg);
#endif
}

static char *find_file(char *dir, int attr, DTA_BUF *dtabuf)
{
	set_dta(dtabuf);	/* setup DTA */
	if (dir != (char *) 0) { /* get first entry */
		reg.h.ah = 0x4E;
		reg.h.cl = (char) attr;
#ifdef LDATA
		reg.x.dx = FP_OFF(dir);
		sreg.ds = FP_SEG(dir);
#else
		reg.x.dx = (unsigned) dir;
#endif
	} else {			/* get next entry */
		reg.h.ah = 0x4F;
	}
#ifdef LDATA
	intdosx(&reg, &nreg, &sreg);
#else
	intdos(&reg, &nreg);
#endif

#ifdef HAVE_CFLAG
	if (nreg.x.cflag)
#else
	if (nreg.x.flags & 1)
#endif
		return (char *) 0;

	if (dtabuf->d_name[0] == 5) dtabuf->d_name[0] = 0xE5;
	return dtabuf->d_name;
}

#else /* !USE_CCL */

static void set_dta(void)
{
	reg.h.ah = 0x1A;
#ifdef LDATA
	reg.x.dx = FP_OFF(&dtabuf);
	sreg.ds = FP_SEG(&dtabuf);
	intdosx(&reg, &nreg, &sreg);
#else
	reg.x.dx = (int) &dtabuf;
	intdos(&reg, &nreg);
#endif
}

static char *find_file(char *dir)
{
	if (dir != (char *) 0) { /* get first entry */
		set_dta();	/* setup DTA */
		reg.h.ah = 0x4E;
		reg.h.cl = ArgsAttr;
#ifdef LDATA
		reg.x.dx = FP_OFF(dir);
		sreg.ds = FP_SEG(dir);
#else
		reg.x.dx = (unsigned) dir;
#endif
	} else {			/* get next entry */
		reg.h.ah = 0x4F;
#ifdef LDATA
		reg.x.dx = FP_OFF(&dtabuf);
		sreg.ds = FP_SEG(&dtabuf);
#else
		reg.x.dx = (unsigned) &dtabuf;
#endif
	}
#ifdef LDATA
	intdosx(&reg, &nreg, &sreg);
#else
	intdos(&reg, &nreg);
#endif

#ifdef HAVE_CFLAG
	if (nreg.x.cflag)
#else
	if (nreg.x.flags & 1)
#endif
		return (char *) 0;

	if (dtabuf.d_name[0] == 5) dtabuf.d_name[0] = 0xE5;
	return dtabuf.d_name;
}

#endif /* USE_CCL */

#endif /* __WIN32__ =================================== */

#if CVLOWER
static char * jstrtolower(char *p)
{
	char *op = p;

	while (*p) {
		if (isk1(*p) && p[1])
			p++;
		else if (('A' <= *p) && (*p <= 'Z'))
			*p |= 0x20;
		p++;
	}
	return op;
}
#else
#define jstrtolower(s)	(s)
#endif

static char * jrindex(char * p, char c)
{
	char           *oldp;

	for (oldp = (char *)0; *p; p++) {
		if (isk1(*p) && isk2(p[1])) {
			p++;
		} else if (*p == c) {
			oldp = p;
		}
	}
	return oldp;
}

#if DO_SORT
static int ustrcmp(const void *s1, const void *s2)
{
	unsigned char * p1 = *(unsigned char **)s1;
	unsigned char * p2 = *(unsigned char **)s2;

	while ((*p1 != '\0') && (*p1 == *p2)) {
		p1++;
		p2++;
	}
	return *p1 - *p2;
}
#endif

static char * xalloc(size_t n)
{
	char  *bp;

	if ((bp = malloc(n)) == (char *) 0)
		error(ERNEM);
	return bp;
}

static char *xrealloc(void *blk, size_t n)
{
	char  *bp;

	if ((bp = realloc(blk, n)) == (char *) 0)
		error(ERNEM);
	return bp;
}

#if USE_CCL
int
expand_arg(char *arg, char *wld,
	char ***argbuf, int *argcnt, size_t *absiz)
{
	char *ap, *qp, *pp, *wp, *cp, *dp;
	int r, l, c, ap_keep;
#ifdef __WIN32__ /* =================================== */
	struct _finddata_t dtabuf;
# define d_attribute attrib		/* <-- CAUTION !, it's dirty */
#else /* __WIN32__ =================================== */
	DTA_BUF    dtabuf;
#endif /* __WIN32__ =================================== */
	static char *work;

	if (wld == NULL) {
		work = xalloc(strlen(arg) + 128 + 1);
		work[0] = '\0';
		wld = work;
	}
	ap = arg;
	qp = wld;
	pp = arg;
	wp = NULL;
	dp = NULL;
	while (*ap != '\0'
		   && (wp == NULL || (*ap != '\\' && *ap != '/'))) {
		if (isk1(*ap) && ap[1] != '\0') {
			*qp++ = *ap++;
			*qp++ = *ap++;
		} else {
			if (wp == NULL
				&& (*ap == '*' || *ap == '?'
					|| *ap == '[' || *ap == LBRACE)) {
				wp = qp;
			} else if (wp == NULL
					   && (*ap == ':' || *ap == '\\' || *ap == '/')) {
				pp = ap + 1;
				wld = qp + 1;
#if !DOT_DLM
				dp = NULL;
			} else if (wp == NULL && *ap == '.') {
#else
			} else if (wp != NULL && *ap == '.') {
#endif
				dp = qp;
			}
			c = *ap++;
			if (isupper(c)) {
				c = tolower(c);
			}
			*qp++ = (char) c;
		}
	}
	*qp = '\0';
	if (wp) {
#if DOT_DLM
		if (dp) {
#else
		if (!dp) {
#endif
			strcpy(wp, "*.*");
		} else {
			strcpy(wp, "*");
		}
	}
	ap_keep = *ap;
	if ((cp = find_file(work, ap_keep ? DirAttr : ArgsAttr, &dtabuf)) != (char *)0) {
		*ap = '\0';
		*wld = '\0';
		do {
			if (*cp != '.' && (r = regcmp(cp, pp)) == 1) {
				if (ap_keep == '\0') {
					strcat(strcpy((*argbuf)[(*argcnt)++] = xalloc(strlen(work) + strlen(cp) + 1), work), jstrtolower(cp));
					if ((size_t) (*argcnt) >= *absiz)
						*argbuf = (char **) xrealloc(*argbuf, (*absiz += REALLOC_UNIT) * sizeof((*argbuf)[0]));
				} else if (dtabuf.d_attribute & DirAttr) {
					strcpy(wld, jstrtolower(cp));
					l = strlen(cp);
					wld[l] = (char) ap_keep;
					wld[l + 1] = '\0';
					if (expand_arg(ap + 1, wld + l + 1, argbuf, argcnt, absiz)) {
						return -1;
					}
				}
			} else if (*cp != '.' && r == -1) {
				return -1;
			}
		} while ((cp = find_file((char *)0, 0, &dtabuf)) != (char *)0);
		*ap = (char) ap_keep;
	}
	return 0;
}
#endif

static void setargv(char * NCptr, int *_ARGC, char ***_ARGV)
{
	int      i;
	char     c;
	char     quote, withwild;	/* flags */
	char    *str;
	char    *ap;
	char    *cp;
	size_t   absiz;
	char    *buf = 0;
	size_t   bfsiz;
#define argbuf (*_ARGV)
#define argcnt (*_ARGC)

#if USE_CCL
	char	*wmk = 0;
	int      no_sort = 0;

	if ((ap = getenv("XARGS")) != NULL
		&& strstr(ap, "NO_SORT") != NULL) {
		no_sort = 1;
	}
	wmk = (char *) xalloc((bfsiz = REALLOC_UNIT * 2) * sizeof(wmk[0]));
#endif /* USE_CCL */

	buf = (char *) xalloc((bfsiz = REALLOC_UNIT * 2) * sizeof(buf[0]));
	ap = _ARGV[0][0];
	argbuf = (char **) xalloc((absiz = REALLOC_UNIT) * sizeof(argbuf[0]));

	for (str = NCptr, i = 0;
	     ((c = *str) != '\r' && c != '\0');
	     str++, i++);
	*str = 0;
	cp = str;
	str = NCptr;
	for (; *str == ' ' || *str == '\t'; str++)
		;
	for (; --cp >= str && (*cp == ' ' || *cp == '\t' || *cp == '\r');)
		;
	cp[1] = 0;
	argbuf[0] = jstrtolower(ap); /* may be my name... */
	argcnt = 1;

	while (*str) {
		withwild = quote = 0;
		while (isspace(*str))
			str++;
		if (*str == 0)
			break;
		ap = buf;
		do {
			if (ap == &buf[bfsiz - 1]) {
#if USE_CCL
				wmk = (char *)xrealloc(buf, (bfsiz + REALLOC_UNIT) * sizeof(wmk[0]));
#endif
				buf = (char *)xrealloc(buf, (bfsiz + REALLOC_UNIT) * sizeof(buf[0]));
				ap = &buf[bfsiz - 1];
				bfsiz += REALLOC_UNIT;
			}
			if (isk1(*str) && str[1]) {
#if USE_CCL
				wmk[ap - buf] = *str;
#endif
				*ap++ = *str++;
				if (ap == &buf[bfsiz - 1]) {
#if USE_CCL
					wmk = (char *)xrealloc(buf, (bfsiz + REALLOC_UNIT) * sizeof(wmk[0]));
#endif
					buf = (char *)xrealloc(buf, (bfsiz + REALLOC_UNIT) * sizeof(buf[0]));
					ap = &buf[bfsiz - 1];
					bfsiz += REALLOC_UNIT;
				}
#if USE_CCL
				wmk[ap - buf] = *str;
#endif
				*ap++ = *str++;
				continue;
			}
#if USE_CCL
			if (!quote && (*str == '*' || *str == '?'
						   || *str == '[' || *str == '{'))
#else
			if (!quote && (*str == '*' || *str == '?'))
#endif /* USE_CCL */
				withwild = 1;
			if (*str == quote) {	/* end of quoted string */
				quote = 0;
				str++;	/* Skip quote */
			} else if (!quote && (*str == '\'' || *str == '"')) {
				quote = *str++;
			} else {
				 if (*str == '\\' && (str[1] == '"' || str[1] == '\''))
					str++;
#if USE_CCL
				 if (*str == '\\' && str[1] == '{') {
					str++;
					wmk[ap - buf] = *str;
				 } else if (!quote && *str == '{') {
					wmk[ap - buf] = LBRACE;
				 } else {
					wmk[ap - buf] = *str;
				 }
#endif
				 *ap++ = *str++;
			}
		} while (*str && (quote || !isspace(*str)));
		*ap = 0;
#if USE_CCL
		wmk[ap - buf] = 0;
		i = argcnt;
		if (withwild) {
			(void) expand_arg(wmk, NULL, &argbuf, &argcnt, &absiz);
		}
		if (argcnt - i > 0) {
#if DO_SORT
			if (!no_sort) {
				qsort(&argbuf[i], argcnt - i, sizeof(argbuf[0]), ustrcmp);
			}
#endif
		} else {
			strcpy(argbuf[argcnt++] = xalloc(strlen(buf) + 1), buf);
			if ((size_t) argcnt >= absiz)
				argbuf = (char **) xrealloc(argbuf, (absiz += REALLOC_UNIT) * sizeof(argbuf[0]));
		}
	}
	free(wmk);
#else /* USE_CCL */
		if (withwild && ((cp = find_file(buf)) != (char *)0)) {
			{
				char * ap1, * ap2;

				ap1 = jrindex(buf, '\\');
				ap2 = jrindex(buf, '/');
				if ((ap = (ap1 < ap2) ? ap2:ap1) == (char *)0)
					ap = jrindex(buf, ':');
			}
			if (ap) {
				ap[1] = 0;
			} else {
				buf[0] = 0;
			}
#if DO_SORT
			i = argcnt;
#endif
			do {
				strcat(strcpy(argbuf[argcnt++] = xalloc(strlen(buf) + strlen(cp) + 1), buf), jstrtolower(cp));
				if (argcnt >= absiz)
					argbuf = (char **) xrealloc(argbuf, (absiz += REALLOC_UNIT) * sizeof(argbuf[0]));
			} while ((cp = find_file((char *)0)) != (char *)0);
#if DO_SORT
			qsort(&argbuf[i], argcnt - i, sizeof(argbuf[0]), ustrcmp);
#endif
		} else {
			argbuf[argcnt++] = ap = xalloc((int) (ap - buf) + 1);
			if (argcnt >= absiz)
				argbuf = (char **) xrealloc(argbuf, (absiz += REALLOC_UNIT) * sizeof(argbuf[0]));
			for (cp = buf; (*ap++ = *cp++) != '\0';)
				;
		}
	}
#endif /* USE_CCL */

	argbuf[argcnt] = (char *) 0;
	argbuf = (char **) xrealloc(argbuf, (argcnt + 1) * sizeof(argbuf[0]));
	free(buf);
}

#if USE_CFG
#define CFG_ALLOC_UNIT	256
#define CFG_REALLOC_UNIT	256
#endif

void
xargs(int *argc, char ***argv)
{
#ifdef __WIN32__ /* =================================== */
	char *pp;
	char s[256];
#else /* __WIN32__ =================================== */
	char far *pp;
	char s[128];
#endif /* __WIN32__ =================================== */
#if USE_CFG
	char *cfg_fname, *cfg_line, *period, *p;
	int cfg_allocated, cfg_used;
	FILE *fp;
#endif
	int  i;

#if USE_CFG
	cfg_fname = xalloc(strlen(*argv[0]) + 4);
	strcpy(cfg_fname, *argv[0]);
	period = NULL;
	for (p = cfg_fname; *p != '\0'; p++) {
		if (isk1(*p) && isk2(p[1])) {
			p++;
		} else if (*p == '\\' || *p == '/') {
			period = NULL;
		} else if (*p == '.') {
			period = p;
		}
	}
	*period = '\0';
	strcat(cfg_fname, ".cfg");
	cfg_line = xalloc(cfg_allocated = CFG_ALLOC_UNIT);
	cfg_line[cfg_used = 0] = '\0';

	if ((fp = fopen(cfg_fname, "r")) != NULL) {
		while (fgets(cfg_line + cfg_used, cfg_allocated - cfg_used, fp) != NULL) {
			cfg_used = strlen(cfg_line);
			if (cfg_line[cfg_used - 1] == '\n') {
				cfg_line[cfg_used - 1] = ' ';
			}
			if (cfg_used + 1 >= cfg_allocated) {
				cfg_line = xrealloc(cfg_line, cfg_allocated += CFG_REALLOC_UNIT);
			}
		}
		fclose(fp);
	}
#endif

#ifdef __WIN32__ /* =================================== */
	pp = &s[0];
	*pp = '\0';

	for(i=1; i<*argc; i++) {
		strcpy(pp, " ");
		pp++;
		strcpy(pp, (*argv)[i]);
		pp += strlen((*argv)[i]);
	}
	i = 0;	/* force next if() false */
#else /* __WIN32__ =================================== */
	pp = MK_FP(PSP, 129);
	for (i = 0; i < 127 && (s[i] = *pp++) != '\0' && s[i] != '\r'; i++)
		;
#endif /* __WIN32__ =================================== */
#if USE_CFG
	s[i] = 0;
	if ((i = cfg_used + 1 + strlen(s) + 1) > cfg_allocated) {
		cfg_line = xrealloc(cfg_line, i);
	}
	strcat(cfg_line, " ");
	strcat(cfg_line, s);
	setargv(cfg_line, argc, argv);
	free(cfg_line);
#else
	if (i == 127 && (s[126] != '\r' && s[126] != 0)) {
		s[126] = 0;
	}
	setargv(s, argc, argv);
#endif
}

#ifdef TEST
#include <stdio.h>
#include <stdlib.h>

static char QUOTE[]  = "\033[33m\"\033[m";

void
main(int argc, char *argv[])
{
	int a = 0;

	(void)setbuf(stdout, malloc(BUFSIZ));	/* for speed up */

	xargs(&argc, &argv);

	printf("[argv]-------------------------- %d argument(s)\n", argc);
	while (argc--) {
		printf("[%4d]%s%s%s\n", a++, QUOTE, *argv++, QUOTE);
	}
	printf("-------------------------------- Thats all\n");
}
#endif
#endif /* __WIN32__ */
